from QueueCatalog import QueueCatalog
from ReplicaCatalog import ReplicaCatalog
from ContentCatalog import ContentCatalog
from LocationClient import LocationClient
from ClaimsClient import ClaimsClient
from ContentClient import ContentClient
from RepositoryClient import RepositoryClient
from Log import Log

import commands
import getopt
import string
import site_conf
import sys
import os
import time
import urllib
from x509 import getX509
import DQ2Mail

x509 = getX509()
if x509 == -1:
    DQ2Mail.sendErrorMail('[BlockVerifier] Proxy certificate expired!')
    sys.exit(1)

class BlockVerifier:
    """Block verifier 'agent'"""

    def __init__(self, logfile):
        self.log = logfile
        self.queue = QueueCatalog(site_conf.queued_transfers_cat, self.log)
        self.content = ContentCatalog(site_conf.local_content_cat)
        self.global_content = ContentClient(x509)
        self.location = LocationClient(x509)
        self.repository = RepositoryClient(x509)
        self.claims = ClaimsClient(proxy_cert=x509)


    def verifyBlock(self):
        """Perform block verification."""

        files = self.queue.findValidatedFiles()

        guids = files.keys()
        
        for guid in guids:
            # ignore if bucket file
            if files[guid]['srcsite'] == site_conf.site_id: continue
            
            # find datasets with vuid
            vuids = self.content.queryDatasetsWithFileByGUID(guid)
            
            for vuid in vuids:
                # delete guid from vuid
                self.log.logDebug( 'Removing '+ guid)
                n = self.content.deleteFileFromDataset(guid, vuid)

                # if was last vuid
                if n == 0:
                    # publish dataset in central
                    self.location.addDatasetReplica(vuid, site_conf.site_id, 'yes')
                    self.log.logInfo("VUID %s published" % vuid)
                    # do callback
                    if site_conf.__dict__.has_key('callback_dataset_complete'):
                        cb = site_conf.callback_dataset_complete
                        cb = cb.replace('[VUID]', vuid)
                        s, o = commands.getstatusoutput("curl -k %s" % cb)
                        if s != 0:
                            self.log.logError('Callback to %s failed!' % cb)
                        else:
                            self.log.logInfo('Callback to %s succeeded!' % cb)

            # delete guid from local queue catalog
            self.queue.deleteFile(guid)
        return 1

    def verifyBucket(self):
        """When the dataset bucket is complete a complete version
        of the dataset is registered in the location catalog. This
        method checks for complete versions of subscriptions and if
        so verifies the content has been copied successfully, then
        deletes the subscription"""

        # get current subscriptions
        subs = self.location.querySubscriptionsInSite(site_conf.site_id)
        
        files = self.queue.findValidatedFiles()

        for vuid in subs.keys():

            self.log.logDebug('Starting to validate vuid '+vuid)
            url = subs[vuid]
            if self.isDatablock(url): continue
 
            if site_conf.site_id in self.location.queryDatasetLocations(vuid, 'yes'):
                # means the ds is complete and the subscription can be removed
                # check all the files have been validated

                filesnotvalidated = []
                guids = self.global_content.queryFilesInDataset(vuid)
                
                # we've lost ds information for buckets so have to go to
                # global catalog. Rather not do this..
                for guid in guids.keys():
                    if guid.upper() not in files.keys():
                        filesnotvalidated.append(guid)
                        self.log.logDebug(guid+' not validated yet')
                    else:
                        self.log.logDebug(guid+' validated')
                    
                if len(filesnotvalidated) == 0:

                    # compare content to remote LRC
                    remote_site = self.getStorage(url)
                    remote_lrc = ReplicaCatalog(site_conf.remote_storages[remote_site][1], self.log)
                    remotefilesvalidated = 1
                    for guid in guids.keys():
                        atts = remote_lrc.getFileAttributes(guid)
                        if atts is None:
                            self.log.logInfo('Validation failed: file '+guid+' is not on remote LRC')
                            remotefilesvalidated = 0
                            break

                    if remotefilesvalidated:
                        self.log.logInfo('Deleting subscription to '+vuid)
                        # delete subs and publish at dest site
                        self.location.deleteDatabucketSubscription(vuid, site_conf.site_id, remote_site)
                        self.location.addDatasetReplica(vuid, remote_site, 'yes')
                        for guid in guids.keys():
                            self.log.logDebug('Deleting file from local catalogs: '+guid)
                            self.queue.deleteFile(guid)
                            # TODO: release file claim (adapt to new Claims code!!)
                            #self.claims.releaseClaim(files[guid]['sourcepfn'])
                                    

    def isDatablock(self, siteId):
        """Check whether this is not a databucket"""
       
        # TODO: Add check to see if siteId URL is valid according to some rules!
  
        return not self.isDatabucket(siteId)

    def isDatabucket(self, url):
        """Check whether this is a site URL plus databucket indication"""
       
        # TODO: Add check to see if siteId URL is valid according to some rules!
  
        d = self.parseQueryString(url)
        if d.has_key('databucket') and d['databucket'][0].upper() == 'Y': return True
        return False

    def getStorage(self, url):
        """Return storage from URL subscription  (e.g '?se=XXX'),
        or blank if no storage indication"""
        d = self.parseQueryString(url)
        if d.has_key('se'): return d['se']
        return ''

    def parseQueryString(self, query):
        """Return a dictionary of key value pairs given in the query string"""
        dict = {}
        params = query.split('?')
        if len(params) != 1:
            params = params[1].split('&')
            for p in params:
                pair = p.split('=')
                pair[1] = urllib.unquote(pair[1])
                #pair[1] = pair[1].replace('%2F', '/')
                dict[pair[0]] = pair[1]
        return dict

    def verify(self):

        self.verifyBucket()
        self.verifyBlock()


def usage():
    """Usage:    [-i nr] [-s secs] [-d secs]

-i: number of iterations
-s: number of seconds
-d: delay between attempts

If -i or -s not specified, stays running as daemon.
"""
    print usage.__doc__


def main(argv):
    
    try:
        opts, args = getopt.getopt(argv, "i:s:d:")
    except getopt.GetoptError:
        print "Invalid arguments!"
        usage()
        sys.exit(2)

    secs = 0
    it = 0
    delay = 20

    for o, a in opts:
        if o == '-s': secs = int(a)
        elif o == '-i': it = int(a)
        elif o == '-d': delay = int(a)
        else:
            usage()
            sys.exit(0)
    if args != []:
        usage()
        sys.exit(0)

    l = Log('BlockVerifier')
    l.logInfo('Starting up..')

    start = int(time.time())
    i = 0
    while 1:
        bv = BlockVerifier(l)
        l.logDebug( 'Looking for verified files to remove')
        bv.verify()
        l.logDebug( 'Done')
 
        if it > 0:
            i += 1
            if i == it: sys.exit(0)
        if secs > 0 and int(time.time())-start > secs:
            sys.exit(0)
        time.sleep(delay)       

if __name__ == '__main__':
    main(sys.argv[1:])
