from QueueCatalog import QueueCatalog
from ReplicaCatalog import ReplicaCatalog
from Log import Log

import GridFTP
import getopt
import string
import site_conf
import sys
import os
import commands
import time
import re


class ReplicaVerifier:
    """Replica verifier 'agent'"""

    def __init__(self, logfile):
        self.log = logfile
        self.childs = []
        self.initCatalogs()

    def initCatalogs(self):
        """Initialize catalogs."""
        
        if self.__dict__.has_key('queue'): del self.queue
        self.queue = QueueCatalog(site_conf.queued_transfers_cat, self.log)

    def countProc(self):
        """Return the current number of verifying processes"""
        (s, o) = commands.getstatusoutput('ps -ax |grep "python ReplicaVerifier.py"|grep -v grep|wc -l')
        if s == 0:
            return int(o.strip())
        return 0

    def verify(self):
        """Perform replica verification for a set of files.
        Compares file size on disk to that in catalog"""

        maxprocs = 50 # hard coded limit
        files = self.queue.findFilesToValidate()
        procs = self.countProc() - 2

        # self.log.logDebug( "Number of processes %d " % procs )
        # self.log.logDebug( "Number of files %d " % len(files) )

        # (s, o) = commands.getstatusoutput('ps -ax |grep "python ReplicaVerifier.py"|grep -v grep')
        # self.log.logDebug( 'ReplicaVerifier processes %s ' % o )

        filesskipvalidation = {}

        if site_conf.__dict__.has_key('ignore_validation_with_fts') and \
            site_conf.ignore_validation_with_fts == 'yes':
            filesskipvalidation = self.queue.findFilesNoValidation()
            files.update(filesskipvalidation)

        for guid in files.keys():

            if procs > maxprocs or procs > len(files):
                self.log.logDebug('Process limit reached')
                return 1

            procs += 1
            pid = os.fork()

            if pid == 0: # child

                os.setsid()
                self.initCatalogs()
                
                if guid not in filesskipvalidation.keys():
                    self.doValidation(guid, files[guid])
                else:
                    self.log.logDebug('No validation required for %s' % files[guid]['pfn'])
                # add the verified files to LRC
                self.addToLRC(guid, files[guid])

                # change state to validated
                r = self.queue.changeState([guid], 'validated')
                if r != []: self.log.logError('Failed to change state of GUID %s' % r)
                sys.exit(0)

            else:
                self.log.logInfo( "Launched worker process %d " % pid )
                self.childs.append( pid )

        return 1


    def doValidation(self, guid, attrs):
        """Do the comparison of attributes of the file on disk and
        those in the catalog"""

        pfn = attrs['pfn']
        self.log.logDebug("About to validate "+pfn)
    
        tool, file = self.resolveProtocolsStat(pfn)
        if tool is None:
            r = self.queue.changeState([guid], 'hold - no way to validate')
            if r != []: self.log.logError('Failed to change state of GUIDs %s' % r)
            sys.exit(0)

        fsize, md5sum = self.statFile(tool, file)

        if fsize < 0:
            r = self.queue.changeState([guid], 'hold - validation failed')
            if r != []: self.log.logError('Failed to change state of GUIDs %s' % r)
            sys.exit(0)

        # for now validates just filesize
        catsize = int(attrs['fsize'])

        # TODO: if catalog size is 0, for now skips comparision...
        if catsize != 0 and catsize != fsize:
            self.log.logError('Validation failed for %s. Catalog: %s, Real: %s. Removing file.' % \
                              (pfn, str(fsize), str(catsize)) )
            tool, file = self.resolveProtocolsRemove(pfn)
            self.removeFile(tool, file)
            # we've lost the old pfn info
            r = self.queue.changeState([guid], 'hold - validation failed')
            if r != []: self.log.logError('Failed to change state of GUIDs %s' % r)
            sys.exit(0)
        else:
            self.log.logInfo('Validated '+pfn)


    def addToLRC(self, guid, attrs):
        """Add the file and attributes to the LRC if it exists"""

        # if it is a bucket
        if attrs['srcsite'] == site_conf.site_id:
            dest_site = attrs['destination'].split('/')[0]
            lrc_endpoint = site_conf.remote_storages[dest_site][1]

            #
            # TODO: Change this policy. Currently deletes source files
            #       if it is a databucket and there is no SOURCE LRC
            #
            if site_conf.local_replica_cat is None:
                self.log.logDebug('Removing source file '+attrs['sourcepfn'])
                tool, file = self.resolveProtocolsRemove(attrs['sourcepfn'])
                self.removeFile(tool, file)

        else: # datablock
            lrc_endpoint = site_conf.local_replica_cat

        if lrc_endpoint is not None:
            lrc = ReplicaCatalog(lrc_endpoint, self.log)
            self.log.logDebug('Adding entry '+guid+' to LRC '+lrc_endpoint)
            res = lrc.addEntry(attrs['pfn'],
                               lfn=attrs['lfn'],
                               guid=guid,
                               filesize=attrs['fsize'],
                               md5sum=attrs['md5sum'])
            if res is None:
                self.log.logError('Error adding file '+attrs['pfn']+' to LRC')
            elif res == 'PFNnonunique':
                self.log.logError('Error adding file '+attrs['pfn']+' to LRC, PFN is not unique')
            else:
                # do callback
                if site_conf.__dict__.has_key('callback_file_copied'):
                    cb = site_conf.callback_file_copied
                    cb = cb.replace('[LFN]', attrs['lfn'])
                    cb = cb.replace('[GUID]', guid)
                    cb = cb.replace('[SURL]', attrs['pfn'])
                    cb = cb.replace('[FSIZE]', attrs['fsize'])
                    cb = cb.replace('[MD5SUM]', attrs['md5sum'])
                    s = commands.getstatus("wget '%s'" % cb)
                    if s != 0:
                        self.log.logError('Callback to %s failed!' % cb)
        else:
            self.log.logInfo('Warning: LRC not available. Registration skipped')



    def getHostname(self, url):
        """Return hostname from URL
        """
        reg = re.search('(?<=://)(\w+\.)*\w+', url)
        host = None
        try:
            host = reg.group(0)
        except:
            # if no storage element, it is a local file
            return ''
        return host


    def resolveProtocolsStat(self, file):
        prot = file.split(':')[0]
        if prot == 'file':
            return 'ls', file[5:]
        elif prot == 'rfio':
            return 'nsls', file[5:]
        elif prot == 'gsiftp':
            return 'gsiftpls', file
        elif prot == 'srm':
            return 'srmls', file
        self.log.logError('Only file, rfio, gsiftp, srm protocols supported at the moment!')
        return None, None
        
    def statFile(self, tool, pfn):
        if tool == 'ls':
            return self.do_ls(pfn)
        elif tool == 'nsls':
            return self.do_nsls(pfn)
        elif tool == 'gsiftpls':
            return self.do_gsiftpls(pfn)
        elif tool == 'srmls':
            return self.do_srmls(pfn)

        self.log.logError('Unsupported tool %s' % str(tool))
        return -1, None

    def do_ls(self, file):
        s, o = commands.getstatusoutput('ls -l %s' % file)
        if s != 0: return -1, None
        return int(o.split()[4]), ''

    def do_nsls(self, file):
        s, o = commands.getstatusoutput('nsls -l %s' % file)
        if s != 0: return -1, None
        return int(o.split()[4]), ''

    def do_srmls(self, file):
        try:
            import SRM
        except:
            self.log.logError("SRM not available in this installation!")
            return -1, None
        try:
            i = file.find('://')+3
            j = file[i:].find('/')
            storage = file[:i+j]
            fm = SRM.srmGetFileMetaData(storage, [file])
            # TODO: Add fm[0].checksumValue after verifying
            #       its type on checksumType
            return fm[0]['size'], ''
        except Exception, e:
            self.log.logError('Error in SRM: %s' % str(e))
            return -1, None

    def do_gsiftpls(self, file):
        try:
            fsize = GridFTP.getFileSize(file)
            return fsize, ''
        except Exception, e:
            self.log.logError('GridFTP error %s' % str(e))
            return -1, None

    def resolveProtocolsRemove(self, file):
        prot = file.split(':')[0]
        if prot == 'file':
            return 'rm', file[5:]
        elif prot == 'rfio':
            return 'rfrm', file[5:]
        elif prot == 'gsiftp':
            return 'gsiftprm', file[7:]
        elif prot == 'srm':
            return 'srmdel', file[4:]
         
        self.log.logError('Only file, rfio protocols supported at the moment!')
        return None, None

    def removeFile(self, tool, pfn):
        if tool == 'rm':
            return self.do_rm(pfn)
        elif tool == 'rfrm':
            return self.do_rfrm(pfn)
        elif tool == 'gsiftprm':
            return self.do_gsiftprm(pfn)
        elif tool == 'srmdel':
            return self.do_srmdel(pfn)
        
        self.log.logError('Unsupported tool %s' % str(tool))
        return False

    def do_rm(self, file):
        s, o = commands.getstatusoutput('rm %s' % file)
        if s != 0:
            self.log.logError('Error removing file %s' % o)
            return False
        return True

    def do_rfrm(self, file):
        s, o = commands.getstatusoutput('rfrm %s' % file)
        if s != 0:
            self.log.logError('Error removing file %s' % o)
            return False
        return True

    def do_gsiftprm(self, file):
        try:
            GridFTP.deleteFile(file)
            return True
        except Exception, e:
            self.log.logError('GridFTP error %s' % str(e))
        return False

    def do_srmdel(self, file):
        # TODO
        print "TODO: Implement SRM delete"


def usage():
    """Usage:    [-i nr] [-s secs] [-d secs]

-i: number of iterations
-s: number of seconds
-d: delay between attempts

If -i or -s not specified, stays running as daemon.
"""
    print usage.__doc__


def main(argv):

    try:
        opts, args = getopt.getopt(argv, "i:s:d:")
    except getopt.GetoptError:
        print "Invalid arguments!"
        usage()
        sys.exit(2)

    secs = 0
    it = 0
    delay = 20

    for o, a in opts:
        if o == '-s': secs = int(a)
        elif o == '-i': it = int(a)
        elif o == '-d': delay = int(a)
        else:
            usage()
            sys.exit(0)
    if args != []:
        usage()
        sys.exit(0)

    l = Log('ReplicaVerifier')
    l.logInfo('Starting up..')
    
    start = int(time.time())
    i = 0
    while 1:
        rv = ReplicaVerifier(l)
        l.logDebug( 'Looking for files to validate')
        rv.verify()

        if it > 0:
            i += 1
            if i == it: sys.exit(0)
        if secs > 0 and int(time.time())-start > secs:
            sys.exit(0)
        time.sleep(delay)

if __name__ == '__main__':
    main(sys.argv[1:])
